import subprocess
import logging
from pathlib import Path
from . import NEDCENC, NEVPK, HEADERFIX
from .constants import SIZE_INFO, VPK_SIZE, VPK

logger = logging.getLogger(__name__)

def _to_bytes(data: bytes | bytearray | str | Path) -> bytes:
    if isinstance(data, (bytes, bytearray)):
        return bytes(data)
    elif isinstance(data, (str, Path)):
        with open(data, "rb") as f:
            return f.read()
    else:
        raise TypeError("Expected bytes, bytearray, or file path (str/Path).")

def _run_tool(command: str):
    result = subprocess.run(
        command,
        capture_output=True,
        text=True,
        shell=True # kept getting FileNotFound for long filenames w/ shell=False + commmand list[]
    )

    if result.stdout:
        for line in result.stdout.splitlines():
            logger.debug(line)

    if result.stderr:
        for line in result.stderr.splitlines():
            logger.error(line)

    if result.returncode != 0:
        raise subprocess.CalledProcessError(result.returncode, command)

def decode_raw(data: bytes | bytearray | str | Path) -> bytearray:
    # handle all input types
    data = _to_bytes(data)
    file = NEDCENC.parent.resolve() / "in.raw"
    logger.debug(f"Writing .raw data to '{file}'.")
    file.write_bytes(data)

    PARENT = NEDCENC.parent.resolve()

    # decode .raw to .bin (includes a vpk compression block which we will extract)
    decoded_path = PARENT / "decoded.bin"
    logger.debug(f"Running nedcenc, output to '{decoded_path}'.")
    _run_tool(f'"{NEDCENC}" -i "{file}" -d -o "{decoded_path}"')
    logger.debug(f"Removing '{file}'.")
    file.unlink()
    if not decoded_path.exists():
        raise Exception("nedcenc did not output a file.")

    # trim the decoded .bin to the vpk block (compressed level data)
    trimmed_path = PARENT / "trimmed.vpk"
    decoded = decoded_path.read_bytes()
    logger.debug(f"Removing '{decoded_path}'.")
    decoded_path.unlink()

    logger.debug("Trimming decoded .raw data to VPK block.")
    size = int.from_bytes(decoded[VPK_SIZE:VPK_SIZE+2], "little")
    trimmed_path.write_bytes(decoded[VPK:VPK+size])

    # decompress the vpk block into pure level data (.bin again)
    decompressed_path = PARENT / "out.bin"
    logger.debug(f"Running nevpk, output to '{decompressed_path}'.")
    _run_tool(f'"{NEVPK}" -i "{trimmed_path}" -d -o "{decompressed_path}"')
    logger.debug(f"Removing '{trimmed_path}'.")
    trimmed_path.unlink()
    if not decompressed_path.exists():
        raise Exception("nevpk did not output a file.")

    data = decompressed_path.read_bytes()
    logger.debug(f"Removing '{decompressed_path}'.")
    decompressed_path.unlink()
    
    logger.info("Conversion from .raw to .bin (decompressed) complete.")
    return bytearray(data)

def encode_raw(original: bytes | bytearray | str | Path,
               data: bytes | bytearray | str | Path) -> bytearray:
    # handle all input types
    original = _to_bytes(original)
    original_file = NEDCENC.parent.resolve() / "original.raw"
    logger.debug(f"Writing .raw original to '{original_file}'.")
    original_file.write_bytes(original)

    data = _to_bytes(data)
    data_file = NEDCENC.parent.resolve() / "in.bin"
    logger.debug(f"Writing .bin data to '{data_file}'.")
    data_file.write_bytes(data)

    PARENT = NEDCENC.parent.resolve()

    # decode original raw to use as template
    raw_decoded_path = PARENT / "raw_decoded.bin"
    logger.debug(f"Running nedcenc, output to '{raw_decoded_path}'.")
    _run_tool(f'"{NEDCENC}" -i "{original_file}" -d -o "{raw_decoded_path}"')
    if not raw_decoded_path.exists():
        raise Exception("nedcenc did not output a file.")
    logger.debug(f"Removing '{original_file}'.")
    original_file.unlink()

    decoded = bytearray(raw_decoded_path.read_bytes())
    logger.debug(f"Removing '{raw_decoded_path}'.")
    raw_decoded_path.unlink()

    # compress new data with nevpk
    compressed_path = PARENT / "trimmed.vpk"
    logger.debug(f"Running nevpk, output to '{compressed_path}'.")
    _run_tool(f'"{NEVPK}" -i "{data_file}" -c -o "{compressed_path}"')
    if not compressed_path.exists():
        raise Exception("nevpk did not output a file.")
    logger.debug(f"Removing '{data_file}'.")
    data_file.unlink()

    vpk = compressed_path.read_bytes()
    logger.debug(f"Removing '{compressed_path}'.")
    compressed_path.unlink()
    
    size = len(vpk)
    if size > 0xFFFF:
        raise ValueError(f"VPK size larger than maximum (expected <= 0xFFFF, got {size:#x}).")

    # re-write size info section
    logger.debug("Using template decoded data to build new deocded data.")
    size_info = int.from_bytes(decoded[SIZE_INFO:SIZE_INFO+4], "little")
    size_info = (size_info & ~(0xFFFF << 9)) | ((size + 2) << 9)
    decoded[SIZE_INFO:SIZE_INFO+4] = size_info.to_bytes(4, "little")
    # re-write vpk size
    decoded[VPK_SIZE:VPK_SIZE+2] = size.to_bytes(2, "little")
    # re-write vpk
    for i in range(size):
        decoded[VPK+i] = vpk[i]
    # write incremenetal padding
    for i in range(VPK + size, 0x530):
        decoded[i] = (i - (VPK + size)) % 0x100
    # write PIKMINCARD 1S ending
    decoded[0x530:0x540] = bytes([
        0x00, 0x00, 0x50, 0x49, 0x4B, 0x4D, 0x49, 0x4E,
        0x43, 0x41, 0x52, 0x44, 0x00, 0x31, 0x53, 0x00
    ])
    
    decoded_path = PARENT / "decoded.bin"
    logger.debug(f"Writing rebuilt decoded data to '{decoded_path}'.")
    decoded_path.write_bytes(bytes(decoded))
    
    # run headerfix to fix the checksum
    # note that headerfix writes the new file in-place
    logger.debug("Running headerfix, output in-place at '{decoded_path}'.")
    _run_tool(f'"{HEADERFIX}" "{decoded_path}"')

    raw_path = PARENT / "card.raw"
    logger.debug(f"Running nedcenc, output to '{raw_path}'.")
    _run_tool(f'"{NEDCENC}" -i "{decoded_path}" -e -o "{raw_path}"')
    logger.debug(f"Removing '{decoded_path}'.")
    decoded_path.unlink()
    
    raw = raw_path.read_bytes()
    logger.debug(f"Removing '{raw_path}'.")
    raw_path.unlink()
    
    logger.info("Conversion from (.bin + .raw) to .raw complete.")
    return bytearray(raw)



from tkinter import Label, messagebox
from tkinterdnd2 import DND_FILES, TkinterDnD
from pathlib import Path
import re
import piket

class ConverterApp:
    def __init__(self, root: TkinterDnD.Tk):
        self.root = root
        self.root.title("Pikmin e+ Converter")
        self.label = Label(
            self.root,
            text="Drag .raw to get .bin (trimmed + decompressed)\n\n" \
                "Drag (.bin + original .raw) to get .raw",
            width=50, height=10, bg="lightgray")
        self.label.pack(padx=10, pady=10)
        self.center_window()

        self.label.drop_target_register(DND_FILES) # type: ignore
        self.label.dnd_bind('<<Drop>>', self.on_file_drop) # type: ignore
        
    def center_window(self):
        root = self.root
        root.update_idletasks()
        width = root.winfo_width()
        height = root.winfo_height()
        screen_width = root.winfo_screenwidth()
        screen_height = root.winfo_screenheight()
        x = (screen_width // 2) - (width // 2)
        y = (screen_height // 2) - (height // 2)
        root.geometry(f"{width}x{height}+{x}+{y}")

    def on_file_drop(self, event):
        paths = [f[0] or f[1] for f in re.findall(r'\{([^}]*)\}|(\S+)', event.data)]
        print(f"Files dropped:\n- {"\n- ".join(paths)}")
        self.process_files([Path(file).resolve() for file in paths])

    def process_files(self, files: list[Path]):
        bin_index = next((i for i, file in enumerate(files) if file.suffix == ".bin"), None)
        if bin_index is not None:
            bin_file = files[bin_index]
            if len(files) != 2 or files[1-bin_index].suffix != ".raw":
                message = "To convert .bin to .raw, you need to ONLY drop both the new .bin and " \
                    "the original .raw (2 files)."
                print(f"[ERROR] {message}")
                messagebox.showerror("Error", message)
                return

            raw_file = files[1-bin_index]
            self.process_bin_file(bin_file, raw_file)

        else:
            for file in files:
                self.process_raw_file(file)
    
    def process_raw_file(self, file: Path):
        out = file.with_suffix(".bin")
        out.write_bytes(piket.decode_raw(file))
        
    def process_bin_file(self, file: Path, original_raw: Path):
        out = original_raw.with_stem(original_raw.stem + "_new")
        out.write_bytes(piket.encode_raw(original_raw, file))

if __name__ == "__main__":
    root = TkinterDnD.Tk()
    app = ConverterApp(root)
    root.mainloop()



import sys
import importlib.resources as resources
from pathlib import Path

_TOOLS = {
    "win32": {
        "libnedclib": "libnedclib.dll",
        "nedcenc": "nedcenc.exe",
        "nevpk": "nevpk.exe",
        "headerfix": "headerfix.exe"
    }
}

# validate platform os support
platform_tools = _TOOLS.get(sys.platform)
if not platform_tools:
    raise OSError(f"Piket currently does not support: {sys.platform}")

# resolve tool paths and expose them
_TOOL_PATHS: dict[str, Path] = {}
for tool, filename in platform_tools.items():
    try:
        with resources.path("piket.bin", filename) as p:
            if not Path(p).exists():
                raise FileNotFoundError(f"Missing required tool: {tool}")
            _TOOL_PATHS[tool] = p
    except Exception as e:
        raise ImportError(f"Error loading binary '{filename}': {e}")

NEDCENC = _TOOL_PATHS["nedcenc"]
NEVPK = _TOOL_PATHS["nevpk"]
HEADERFIX = _TOOL_PATHS["headerfix"]

# expose functions
from .main import decode_raw, encode_raw

__all__ = ["NEDCENC", "NEVPK", "HEADERFIX"]
